# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/attribute.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/attribute.proto\x12\x0fsarus_data_spec\"\xaa\x01\n\tAttribute\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0e\n\x06object\x18\x02 \x01(\t\x12>\n\nproperties\x18\x03 \x03(\x0b\x32*.sarus_data_spec.Attribute.PropertiesEntry\x12\x0c\n\x04name\x18\x04 \x01(\t\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.attribute_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ATTRIBUTE_PROPERTIESENTRY._options = None
  _ATTRIBUTE_PROPERTIESENTRY._serialized_options = b'8\001'
  _ATTRIBUTE._serialized_start=62
  _ATTRIBUTE._serialized_end=232
  _ATTRIBUTE_PROPERTIESENTRY._serialized_start=183
  _ATTRIBUTE_PROPERTIESENTRY._serialized_end=232
# @@protoc_insertion_point(module_scope)
